/**
 * Title:	    Traces
 * Copyright:	Copyright (c) Jan-Pascal van Best
 * Email:	    janpascal@vanbest.org
 * Author:	    Jan-Pascal van Best
 * Web: 	    http://www.tbm.tudelft.nl/webstaf/janb/software/traces/api/index.html
 **************************************************************************
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation, version 2.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

package nl.tudelft.tbm.matrix;


// Standard Java packages
import java.util.*;
import java.util.logging.*;
import java.lang.*;

public class NNLS {
    /** 
     * Solves the matrix equation Ax=b in the least-squares sense,
     * taking into account the constraint x>=0. Algorithm from
     * Charles L. Lawson, Richard J. Hanson, Solving Least Squares Problems, 
     * SIAM, Philadelphia, 1995
     * A is an mxn matrix of doubles; b is an array of m doubles
     */

    private static Logger logger = Logger.getLogger("nl.tudelft.tbm.matrix");
    public static double PRECISION = 1E-20;

    public static Matrix NNLS(Matrix AA, Matrix BB) {
        SimpleMatrix A;
        SimpleMatrix B;
        if (AA instanceof SimpleMatrix) {
            A = (SimpleMatrix) AA;
        } else {
            A = new SimpleMatrix(AA);
        }

        if (BB instanceof SimpleMatrix) {
            B = (SimpleMatrix) BB;
        } else {
            B = new SimpleMatrix(BB);
        }

        SimpleMatrix AT = (SimpleMatrix) A.transpose();

        int m = A.getRows();
        int n = A.getColumns();

        boolean[] P = new boolean[n];
        boolean[] Z = new boolean[n];
        for(int i=0; i<n; i++) {
            P[i] = false;
            Z[i] = true;
        }
        int Zsize = n;

        SimpleMatrix x = new SimpleMatrix(n,1);
//        for(int i=0; i<n; i++) { x.a[i][0] = 0; };

        for (;;) { // outer loop
            // Step 2
            SimpleMatrix w = (SimpleMatrix) AT.times(B.minus(A.times(x)));
//            logger.info("w:");
//            logger.info(w.toString(8,3));

            // Step 3
            boolean finished = true;
            for(int i=0; i<n && finished; i++) {
                if (Z[i] && w.a[i][0]>0) finished=false; // Was: w.a[i][0]>0
            }
            if (finished) {
                logger.info("Finished outer loop, Z;w: ");
                for(int i=0; i<n; i++) {
                    logger.info(Z[i]+"  "+w.a[i][0]);
                }
                break; // Complete algorithm completed
            }

            // Step 4
            int t=0;
            double wt = 0; // There is at least one t such that Z[t] and w.a[t]>0
            for(int i=0; i<n; i++) {
                if (Z[i] && w.a[i][0]>wt) {
                    t = i;
                    wt = w.a[i][0];
                }
            }

            logger.info("t="+t+"; wt="+wt);
            // Step 5
            Z[t] = false;
            P[t] = true;
            Zsize--;

            for (;;) { // inner loop
                // Step 6
                SimpleMatrix Ap = new SimpleMatrix(m,n);
                for(int j=0; j<n; j++) {
                    // default value is zero
                    if(!Z[j]) {
                        for (int i=0; i<m; i++) Ap.a[i][j] = A.a[i][j];
                    }
                }

//                System.out.println("Ap:");
//                Ap.print(6,2);

                // Step 6b. FIXME: VERY INEFFICIENT because of copying of matrices
                // Also doesn't work, because the Ap are clearly not full rank,
                // which this QR decomposition requires
//                Jama.QRDecomposition qr = new Jama.QRDecomposition(simpleToJama(Ap));
//                SimpleMatrix z = jamaToSimple(qr.solve(simpleToJama(B)));
                // This overwrites Ap
                SimpleMatrix z = new SimpleMatrix(m,1);
                for(int j=0; j<m; j++) {
                    z.a[j][0] = B.a[j][0];
                }
                int[] JPVT = new int[n];
//                for(int j=0; j<n; j++) {
//                    JPVT[j] = Z[j]?0:1;
//                }
                org.netlib.util.intW RANK = new org.netlib.util.intW(0);
                double[] work = new double[(Math.max( Math.min(m,n)+3*n, 2*Math.min(m,n)+1 ))];
                org.netlib.util.intW INFO = new org.netlib.util.intW(0);
                org.netlib.lapack.DGELSX.DGELSX(m, n, 1, Ap.a, z.a, JPVT, 1E-15, RANK, work, INFO);

                logger.info("RANK: "+RANK.val);
//                logger.info("INFO: "+INFO.val);
                if(INFO.val<0) logger.severe(INFO.val+"th argument had an illegal value");

//                logger.info("z:");
//                for(int i=0; i<n; i++) { logger.info(""+z.a[i][0]); }

                // Step 6c
                for(int j=0; j<n; j++) {
                    // default value is zero
                    if(Z[j]) {
                        for (int i=0; i<m; i++) z.a[j][0] = 0;
                    }
                }

//                System.out.println("z (filled zeros where appropriate):");
//                z.print(8,3);

                // Step 7
                finished = true;
                for(int i=0; i<n && finished; i++) {
                    if (P[i] && z.a[i][0]<=0) finished=false;
                }
                if (finished) {
                    logger.info("Finished inner loop");
                    for(int i=0; i<n; i++) {
                        x.a[i][0] = z.a[i][0];
                    }
                    break; // Goto step 2: Completed inner loop, start again with outer loop
                }

                // Step 8,9
                int q=0;
                double alpha = Double.MAX_VALUE;
                for(int i=0; i<n; i++) {
                    if (P[i] && z.a[i][0]<=0) {
                        double value = x.a[i][0]/(x.a[i][0]-z.a[i][0]);
                        if (value<alpha) {
                            alpha = value;
                            q = i;
                        }
                    }
                }

                // Step 10
                for(int i=0; i<n; i++) {
                    x.a[i][0] += alpha*(z.a[i][0]-x.a[i][0]);
                }

//                logger.info("New x:");
//                logger.info(x.toString(8,3));

                // Step 11
                for(int i=0; i<n; i++) {
                    if (P[i] && x.a[i][0]==0.0) { // Was: x.a[i][0]==0.0
                        P[i] = false;
                        Z[i] = true;
                    }
                }

                // Goto step 6

            } // End of inner loop

        }

        return x;
    }
/*     
    static Jama.Matrix simpleToJama(SimpleMatrix A) {
        return Jama.Matrix.constructWithCopy(A.getArray());
    }

    static SimpleMatrix jamaToSimple(Jama.Matrix A) {
        return new SimpleMatrix(A.getArray());
    }
*/
}
