/**
 * Title:	Traces
 * Copyright:	Copyright (c) Jan-Pascal van Best
 * Email:	    janpascal@vanbest.org
 * @author	    Jan-Pascal van Best
 **************************************************************************
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation, version 2.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

package nl.tudelft.tbm.matrix;


import java.util.*;
import java.io.*;
import java.lang.*;
import java.text.*;

public class SimpleMatrix implements Matrix {
	int m;
	int n;

	double[][] a;
    int[] pivots;
    boolean reduced;
    int rank;

	public SimpleMatrix(int m, int n) {
		this.m = m;
		this.n = n;
		a = new double[m][n];
	}

	public SimpleMatrix(double[][] A) {
		m = A.length;
		if (m==0) {
			n = 0;
		} else {
			n = A[0].length;
		}

		a = new double[m][n];

		for(int i=0; i<m; i++) {
			for(int j=0; j<n; j++) {
				a[i][j] = A[i][j];
			}
		}
	}

	public SimpleMatrix(double[] A) {
        m = 1;
		n = A.length;
		a = new double[m][n];

        for(int j=0; j<n; j++) {
            a[0][j] = A[j];
        }
	}

	public SimpleMatrix(Matrix other) {
        m = other.getRows();
        n = other.getColumns();
		a = new double[m][n];

		for(int i=0; i<m; i++) {
			for(int j=0; j<n; j++) {
				a[i][j] = other.get(i,j);
			}
		}
	}

    public int getRows() { return m; }
    public int getColumns() { return n; }

	public double get(int i, int j) {
		return a[i][j];
	}

	public void set(int i, int j, double value) {
		a[i][j] = value;
    }

    public double[][] getArray() {
        return a;
    }

	public Matrix transpose() {
        SimpleMatrix result = new SimpleMatrix(n,m);
		for(int i=0; i<m; i++) {
			for(int j=0; j<n; j++) {
				result.a[j][i] = a[i][j];
			}
		}
          
		return result;
	}

	// matrix algebra
	public Matrix times(Matrix other) {
        int o = other.getColumns();
        SimpleMatrix result = new SimpleMatrix(m,o);

		for(int i=0; i<m; i++) {
			for(int j=0; j<o; j++) {
                double v = 0;
    			for(int k=0; k<n; k++) {
                    v += a[i][k] * other.get(k,j);
                }
                result.a[i][j] = v;
            }
        }

        return result;
    }

	public Matrix times(double other) {
        SimpleMatrix result = new SimpleMatrix(m,n);
		for(int i=0; i<m; i++) {
			for(int j=0; j<n; j++) {
				result.a[i][j] = other * a[i][j];
			}
		}
          
		return result;
    }

	public Matrix plus(Matrix other) {
        SimpleMatrix result = new SimpleMatrix(m,n);
		for(int i=0; i<m; i++) {
			for(int j=0; j<n; j++) {
				result.a[i][j] = a[i][j] + other.get(i,j);
			}
		}
          
		return result;
    }

	public Matrix minus(Matrix other) {
        SimpleMatrix result = new SimpleMatrix(m,n);
		for(int i=0; i<m; i++) {
			for(int j=0; j<n; j++) {
				result.a[i][j] = a[i][j] - other.get(i,j);
			}
		}
          
		return result;
    }

	// elementary operations
	public void swapRow(int r1, int r2) {
        double temp;
        for(int j=0; j<n; j++) {
            temp = a[r1][j];
            a[r1][j] = a[r2][j];
            a[r2][j] = temp;
        }
//        System.out.println("Swap ("+r1+","+r2+")");
//        print(5,1);
    }

	public void scaleRow(int row, double factor) {
        for(int j=0; j<n; j++) {
            a[row][j] = factor * a[row][j];
        }
//        System.out.println("Scale ("+row+","+factor+")");
//        print(5,1);
    }
	// r1 := r1 + factor * r2
	public void addRow(int r1, int r2, double factor) {
        for(int j=0; j<n; j++) {
            a[r1][j] = a[r1][j] + factor * a[r2][j];
        }
//        System.out.println("Add ("+r1+":="+r1+"+"+factor+"*"+r2+")");
//        print(5,1);
    }

	//
    public int rank() {
        return 0;
    }

    int gaussJordan() {
        return gaussJordan(m);
    }

    int gaussJordan(int maxRows) {
        pivots = new int[m];

        // First sweep the matrix
//        System.out.print("Sweeping, i=");
        int pivot = 0;
        int i;
        for(i=0; i<maxRows; i++){
//            System.out.print(" "+i);
            // swap rows until we have a pivot
            while (pivot<n && a[i][pivot] == (double) 0) {
                for(int k=i; k<maxRows; k++) {
                    if (a[k][pivot] != (double) 0) {
                        swapRow(i,k);
                        break;
                    }
                }
                if(a[i][pivot] == (double) 0) {
                    pivot++;
                }
            }
//            System.out.println("pivot: (p="+pivot+")");
            pivots[i] = pivot;

            if(pivot==n) {
                // finished
                break;
            }

            // get leading one
            scaleRow(i, 1.0/a[i][pivot]);

            //get zeros above leading one
            for(int k=i-1; k>=0; k--)
                addRow(k,i,-1.0*a[k][pivot]);

            //get zeros below leading one
            for(int k=i+1; k<maxRows; k++)
                addRow(k,i,-1.0*a[k][pivot]);

        } // end for loop

        rank = i;

        // Fill in rest of pivots
        for (int k=i; k<m; k++) pivots[k] = n;

        // Clean up "-0"s
        for (i=0; i<maxRows; i++) {
            for (int j=0; j<n; j++) {
                if (a[i][j] == (double) -0.0) a[i][j] = (double) 0;
            }
        }

        return rank;
                              
//        System.out.println("Rank: "+rank);
    }

    public Matrix inverse() {
        if (n!=m) return null;
//        System.out.println("Calculating inverse...");
        SimpleMatrix temp = new SimpleMatrix(m, 2*m);
        for(int i=0; i<m; i++) {
            for(int j=0; j<m; j++) {
                temp.a[i][j] = a[i][j];
            }
            temp.a[i][i+m] = 1;
        }
//        System.out.println("Reducing:");
//        temp.print(6,2);
        if (temp.gaussJordan()!=m) {
//            temp.print(6,2);
//            System.out.println("Failed");
            return null;
        }
//        temp.print(6,2);
        SimpleMatrix result = new SimpleMatrix(m,m);
        for(int i=0; i<m; i++) {
            for(int j=0; j<m; j++) {
                result.a[i][j] = temp.a[i][j+m];
            }
        }
        return result;
    }

    public String toString(int width, int decimals) {
        StringBuffer s = new StringBuffer(m*n*width);
        DecimalFormat format = new DecimalFormat();
        format.setDecimalFormatSymbols(new DecimalFormatSymbols(Locale.US));
        format.setMinimumIntegerDigits(1);
        format.setMaximumFractionDigits(decimals);
        format.setMinimumFractionDigits(decimals);
        format.setGroupingUsed(false);
        for(int i=0; i<m; i++) {
            for (int j=0; j<n; j++) {
                String number = format.format(a[i][j]);
                int padding = Math.max(1,width-number.length());
                s.append("                                   ".substring(0,padding));
                s.append(number);
            }
            s.append('\n');
        }
        return s.toString();
    }

   public void print (int w, int d) {
       System.out.print(toString(w,d));
   }
}

