/**
 * Title:	Traces
 * Copyright:	Copyright (c) Jan-Pascal van Best
 * Email:	    janpascal@vanbest.org
 * @author	    Jan-Pascal van Best
 **************************************************************************
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation, version 2.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

package nl.tudelft.tbm.matrix;


import java.util.*;
import java.io.*;
import java.lang.*;
import java.text.*;
import java.util.logging.*;

public class VectorSpace {
    // A contains basis vectors in rows
    // A has an extra row to facilitate testing
    // whether a vector fits in the vector space
    SimpleMatrix A;
    double[][] a; 
    int[] pivots;
	int m;
	int n;
    // N contains null space vectors
    SimpleMatrix N;

    private static Logger logger = Logger.getLogger("nl.tudelft.tbm.matrix");

    private static DecimalFormat format;
    {   format = new DecimalFormat();
        format.setDecimalFormatSymbols(new DecimalFormatSymbols(Locale.US));
        format.setMinimumIntegerDigits(1);
        format.setMaximumFractionDigits(0);
        format.setMinimumFractionDigits(0);
        format.setGroupingUsed(false);
    }

    private VectorSpace(int reserveSpaceForVectors, int n) {
        A = new SimpleMatrix(1+reserveSpaceForVectors, n);
        a = A.a;
        this.m = 0;
        this.n = n;
        pivots = new int[0];
    }

	public VectorSpace(Matrix A) {
        SimpleMatrix copy = new SimpleMatrix(A);
        copy.gaussJordan();
        m = copy.rank;
        n = copy.n;
        this.A = new SimpleMatrix(m+1,n);
        a = this.A.a;
        pivots = copy.pivots;
        for(int i=0; i<m; i++) {
            for (int j=0; j<n; j++) {
                a[i][j] = copy.a[i][j];
            }
        }

        StringBuffer s = new StringBuffer("Pivots: ");
        for(int i=0; i<pivots.length; i++) s.append(pivots[i]+" ");
        logger.finer(s.toString());
/*
		logger.info("Calculating basis of null space...");
        // Now create basis of null space
        // See Fraleigh&Beauregard, Linear Algebra, 1987,
        // Section 3.5 for algorithm

        int nulldim = n-m;
        // temp will contain null space basis vectors in certain _columns_
        SimpleMatrix temp = new SimpleMatrix(n, n);
        int l = 0; // counts elements of the pivots array

        boolean[] where = new boolean[n];

        {
            int i=0; // i,j are indices to the copy[][] array, contains the GJ form of A
            int j=0;
            do {
                if (pivots[l]==j) {
                    // pivot column, copy relevant entries into null space basis
                    for(int k=j+1; k<n; k++) {
                        temp.set(j,k, -copy.get(i,k));
                    }
                    where[j] = false;
                    i++;
                    j++;
                    l++;
                } else {
                    // non-pivot column
                    i--; // go back up 
                    temp.set(j,j, 1);
                    where[j] = true;
                    i++;
                    j++;
                }
            } while (j<n);
        }

        N = new SimpleMatrix(nulldim, n);
        int k = -1;
        for(int i=0; i<nulldim; i++) {
            // find which vector to copy from the temp matrix, using the where[] array
            do {
                k++;
            } while (!where[k]);
            for(int j=0; j<n; j++) {
                N.set(i,j, temp.get(j,k));
            }
        }
*/
	}

    // Only for the original vectorspace, does not take into account
    // vectors added to the space later
    public Matrix getNullBasis() {
        return N;
    }

    public Matrix getBasis() {
        return A;
    }

    public void print(int w, int d) {
        A.print(w,d);
    }

    public String toString(int width, int decimals) {
        return A.toString(width, decimals);
    }

    public int getDimension() {
        return m;
    }

    // v should be a 1 by n matrix
    public void add(SimpleMatrix v) {
        add(v.a[0]);
    }

    // v should be a 1 by n matrix
    public void add(double[] v) {
        if (A.getRows() < m+2) {
            SimpleMatrix copy = new SimpleMatrix(m+2, n);
            for(int i=0; i<m; i++) {
                for (int j=0; j<n; j++) {
                    copy.a[i][j] = a[i][j];
                }
            }
            this.A = copy;
            this.a = copy.a;
        }
        for (int j=0; j<n; j++) {
            A.a[m][j] = v[j];
        }
        A.gaussJordan(m+1); // GJ the first m+1 rows,to see if the new vectors's really new
        pivots = A.pivots;
        if (A.rank > m) {
            m = A.rank;
        }
    }

    // v should be a 1 by n matrix
    public boolean contains(SimpleMatrix v) {
        return contains(v.a[0]);
    }

    // v should be a 1 by n matrix
    public boolean contains(double[] v) {
        for(int j=0; j<n; j++) a[m][j] = v[j];
        for(int i=0; i<m; i++){
            A.addRow(m, i, -a[m][pivots[i]]);
        }

        boolean result = true;
        for(int j=0; j<n; j++) {
            if (a[m][j] != (double) 0) {
                result = false;
                break;
            }
        }
        return result;
    }

    // Generates a "nice" basis of the vector space. In the
    // returned matrix, these vectors are in the columns.
    // The nice basis 
    // - Has only vectors with only ones and zeros
    // - Consists of subvectors of the rows of B
    // - Has as few ones in the constituent vectors as possible
	public SimpleMatrix makeNiceBasis(SimpleMatrix B) {
        SimpleMatrix vectors = new SimpleMatrix(n,m);
		VectorSpace result = new VectorSpace(n, n);

        int cutoff = cutoffOnes(B);

        int found = 0;
		for(int k=1; k<cutoff && found<m; k++) {
            logger.fine("Trying vectors with "+k+" ones");
            found = makeBasisStepBruteForce(vectors, result, k, found);
        }

		for(int k=cutoff; k<=n && found<m; k++) {
            logger.fine("Trying vectors with "+k+" ones");
            found = makeBasisStepFromRows(B, vectors, result, k, found);
        }

        return vectors;
    }

    // Same as makeNiceBasis, but include basis for null space
    // at the end of the basis.
	public SimpleMatrix makeNiceBasisIncludeNull(SimpleMatrix B) {
        SimpleMatrix vectors = new SimpleMatrix(n,n);
		VectorSpace result = new VectorSpace(n, n);

        int cutoff = cutoffOnes(B);

        int found = 0;
		for(int k=1; k<cutoff && found<m; k++) {
            logger.fine("Trying vectors with "+k+" ones");
            found = makeBasisStepBruteForce(vectors, result, k, found);
        }

		for(int k=cutoff; k<=n && found<m; k++) {
            logger.fine("Trying vectors with "+k+" ones");
            found = makeBasisStepFromRows(B, vectors, result, k, found);
        }

        for(int j=m; j<n; j++) {
            for(int i=0; i<n; i++) {
                vectors.a[i][j] = N.a[j-m][i];
            }
        }

        return vectors;
    }

    static int cutoffOnes(SimpleMatrix A) {
        int m = A.getRows();
        int n = A.getColumns();

        int[] onesInRow = new int[m];
        for(int i=0; i<m; i++) {
            int count = 0;
            for(int j=0; j<n; j++) {
                if (A.get(i,j)>0.5) count++;
            }
            onesInRow[i] = count;
        }
        int k=0;
        double doAllMethod;
        double otherMethod;
        do {
            k++;
            // Simple method examines all vectors of length n
            // with k ones
            doAllMethod = Functions.nCr(n, k);

            // Other method takes all row vectors in A and 
            // examines the vectors with k of the ones in each row vector
            otherMethod = 0;
            for (int i=0; i<m; i++) {
                if (k<=onesInRow[i]) {
                    otherMethod += Functions.nCr(onesInRow[i], k);
                }
            }
            logger.fine("k="+k+" would mean "+doAllMethod+" operations with the simple method and "+otherMethod+" with the smart method.");
        } while (doAllMethod < otherMethod);

        return k;
    }

    int makeBasisStepBruteForce(SimpleMatrix vectors, VectorSpace result, int k, int found) {
        // Generate all vectors with k ones, rest zeros
        // First make a vector with the first k coefficients one, rest zero
        double[] v = new double[n];
        for (int j=0; j<k; j++) {
            v[j] = 1.0;
        }

        // Then transform this vector into all the others
        int count = 0;
        do {
            if(count%50 == 0) {
                logger.finer("Trying "+vectorToString(v));
            }
            if (contains(v) && !result.contains(v)) {
                result.add(v);
                for(int j=0; j<n; j++) {
                    vectors.a[j][found] = v[j];
                }
                found++;
//                logger.info("Basis is now: ");
//                vectors.print(2,0);
            }
            count++;
        } while (nextBasisTryBruteForce(v) && found<m);
        return found;
    }
/*
    public SimpleMatrix makeNiceBasisBruteForce() {
        SimpleMatrix vectors = new SimpleMatrix(m,n);
		VectorSpace result = new VectorSpace(n);

        int found = 0;
		for(int i=1; i<=n && found<m; i++) {
			logger.info("Trying vectors with "+i+" ones");
			// Generate all vectors with i ones, rest zeros
			// First make a vector with the first i coefficients one, rest zero
			double[] v = new double[n];
			for (int j=0; j<i; j++) {
				v[j] = 1.0;
			}

			int count = 0;
			do {
				if(count%500 == 0) {
                    StringBuffer s = new StringBuffer(n*4);
                    for(int j=0; j<n; j++) {
                        s.append(" "+v[j]);
                    }
                    logger.info("Vector: "+s);
                }
				if (contains(v) && !result.contains(v)) {
//					logger.info("added vector to basis:");
//					v.print(1,0);
					result.add(v);
                    for(int j=0; j<n; j++) {
                        vectors.a[found][j] = v[j];
                    }
                    found++;
                    logger.info("Basis is now: ");
                    vectors.print(2,0);
				}
				count++;
			} while (nextBasisTryBruteForce(v) && found<m);
		}

		return vectors;
    }
*/
    public static boolean nextBasisTryBruteForce(double[] v) {
		int numberOfOnesHad = 0;
		int cols = v.length;
		boolean finished = false;
		boolean noMore = false;
		int lastOne = cols-1;
		while (lastOne>0 && v[lastOne]<1) lastOne--;
		if (lastOne!=cols-1) {
			v[lastOne] = 0;
			v[lastOne+1] = 1;
		} else {
			do {
				v[lastOne] = 0;
				numberOfOnesHad++;
				int previousOne = lastOne-1;
				while (previousOne>=0 && v[previousOne]<0.5) previousOne--;
				if (previousOne<0) {
					return false;
				}
				if (previousOne!=lastOne-1) {
					v[previousOne] = 0;
					for(int j=0; j<numberOfOnesHad+1; j++) {
						v[previousOne+1+j] = 1;
					}
					finished = true;
				} else {
					lastOne = previousOne;
				}
			} while (!finished);
		}
		return !noMore;
	}

    int makeBasisStepFromRows(SimpleMatrix B, SimpleMatrix vectors, VectorSpace result, int k, int found) {
        int count = 0;
        for(int i=0; i<B.getRows(); i++) {
            // Try vectors with k ones from row i
            double[] row = new double[n];
            for(int j=0; j<n; j++) {
                row[j] = B.getArray()[i][j];
            }
            double[] v = new double[n];
            int place = -1;
            boolean finishedRow = false;
            for (int j=0; j<k; j++) {
                place = findNextOne(row,place+1);
                if(place>=n) {
                    finishedRow = true;
                    break;
                }
                v[place] = (double) 1;
            }
            if(finishedRow) continue;

            logger.finer("Looking for "+k+" ones in row "+i); // +": "+vectorToString(v));

            do {
//                logger.info("Trying: "+vectorToString(v));
                if (contains(v) && !result.contains(v)) {
					logger.finer("Found "+found+"th nice basis vector");
//					v.print(1,0);
                    result.add(v);
                    for(int j=0; j<n; j++) {
                        vectors.a[j][found] = v[j];
                    }
                    found++;
//                        logger.info("Basis is now: ");
//                        vectors.print(5,2);
                }
                count++;
            } while (nextBasisTryFromRow(row, v) && found<m);
        }
        logger.fine("Examined "+count+" vectors with "+k+" ones.");
        return found;
    }
/*
    public SimpleMatrix makeNiceBasisFromRows(SimpleMatrix B) {
        logger.info("Getting rows from matrix:");
        B.print(1,0);

        SimpleMatrix vectors = new SimpleMatrix(m,n);
		VectorSpace result = new VectorSpace(n);

        int found = 0;
		for(int k=1; k<=n && found<m; k++) {
			logger.info("Trying vectors with "+k+" ones");
            for(int i=0; i<B.getColumns(); i++) {
                // Try vectors with k ones from row i
                double[] row = new double[n];
                for(int j=0; j<n; j++) {
                    row[j] = B.a[i][j];
                }
    			double[] v = new double[n];
                int place = -1;
                boolean finishedRow = false;
    			for (int j=0; j<k; j++) {
                    place = findNextOne(row,place+1);
                    if(place>=n) {
                        finishedRow = true;
                        break;
                    }
                    v[place] = (double) 1;
    			}
                if(finishedRow) continue;

                logger.info("Looking for "+k+" ones in row "+i); // +": "+vectorToString(v));

                int count = 0;
                do {
//                    if(count%25 == 0) {
//                        logger.info("Trying: "+vectorToString(v));
//                    }
                    if (contains(v) && !result.contains(v)) {
    //					logger.info("added vector to basis:");
    //					v.print(1,0);
                        result.add(v);
                        for(int j=0; j<n; j++) {
                            vectors.a[found][j] = v[j]/k;
                        }
                        found++;
//                        logger.info("Basis is now: ");
//                        vectors.print(5,2);
                    }
                    count++;
                } while (nextBasisTryFromRow(row, v) && found<m);
            }
        }

		return vectors;
    }
*/
    final static boolean nextBasisTryFromRow(double[] row, double[] v) {
		int cols = v.length;
		boolean noMore = false;
		int lastOne = findPreviousOne(v, cols-1);  // Last one in 1
        int nextOne = findNextOne(row, lastOne+1); // First available one in row after lastOne
		if (nextOne<cols) {
			v[lastOne] = 0;
			v[nextOne] = 1;
		} else {
    		int numberOfOnesHad = 0;
    		boolean finished = false;
			do {
				v[lastOne] = 0;
				numberOfOnesHad++;
				int previousOne = findPreviousOne(v, lastOne-1);
				if (previousOne<0) {
					return false;
				}
				if (previousOne!=findPreviousOne(row, lastOne-1)) {
					v[previousOne] = 0;
                    nextOne = findNextOne(row, previousOne+1);
					for(int j=0; j<numberOfOnesHad+1; j++) {
                        if (nextOne>=row.length) {
                            return false;
                        }
                        v[nextOne] = (double) 1;
                        nextOne = findNextOne(row, nextOne+1);
					}
					finished = true;
				} else {
					lastOne = previousOne;
				}
			} while (!finished);
		}
		return !noMore;
	}

    final static int findNextOne(double[] row, int start) {
        int i;
        for(i=start; i<row.length && row[i]<0.5; i++);
        return i;
    }

    final static int findPreviousOne(double[] row, int start) {
        int i;
        for(i=start; i>=0 && row[i]<0.5; i--);
        return i;
    }

    final static String vectorToString(double[] v) {
        StringBuffer s = new StringBuffer(v.length*4);
        for(int j=0; j<v.length; j++) {
            s.append(" ");
            s.append(format.format(v[j]));
        }
        return s.toString();
    }
}
